// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "Stage.h"
#include "MainFrm.h"
#include "indevdlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNCREATE(CMainFrame, CFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ACTIVATEAPP()
	ON_WM_PALETTECHANGED()
	ON_WM_MOVE()
	ON_COMMAND(ID_FILE_NEW, OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_EDIT_INSERTSHAPE, OnEditInsertShape)
	ON_COMMAND(ID_EDIT_TRANSFORM, OnEditTransform)
	ON_UPDATE_COMMAND_UI(ID_EDIT_TRANSFORM, OnUpdateEditTransform)
	ON_WM_SETFOCUS()
	ON_COMMAND(ID_EDIT_CONTROL, OnEditControl)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CONTROL, OnUpdateEditControl)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SELECTION, OnUpdateViewSelection)
	ON_COMMAND(ID_VIEW_SELECTION, OnViewSelection)
	ON_UPDATE_COMMAND_UI(ID_EDIT_DELETESHAPE, OnUpdateEditDeleteshape)
	ON_COMMAND(ID_EDIT_DELETESHAPE, OnEditDeleteshape)
	ON_COMMAND(ID_EDIT_INPUTDEV, OnEditInputdev)
	ON_UPDATE_COMMAND_UI(ID_VIEW_HITPT, OnUpdateViewHitpt)
	ON_COMMAND(ID_VIEW_HITPT, OnViewHitpt)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
	m_pScene = NULL;
	m_bRotate = FALSE;
	m_pCurShape = FALSE;
	m_pTransDlg = NULL;
	m_pController = NULL;
	m_iObjCtrlDev = 0;
	m_iControlObj = 0;
	m_iControlType = 0;
	m_pSelBox = NULL;
	m_iHitFace = -1;
	m_pHitPtr = NULL;
}

CMainFrame::~CMainFrame()
{
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	if (!m_wndToolBar.Create(this) ||
		!m_wndToolBar.LoadToolBar(IDR_MAINFRAME))
	{
		TRACE0("Failed to create toolbar\n");
		return -1;      // fail to create
	}

	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

	// Create the 3D window
	if (!m_wnd3d.Create(this, IDC_3DWND)) {
		return -1;
	}

	// TODO: Remove this if you don't want tool tips or a resizeable toolbar
	m_wndToolBar.SetBarStyle(m_wndToolBar.GetBarStyle() |
		CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC);

	// TODO: Delete these three lines if you don't want the toolbar to
	//  be dockable
	m_wndToolBar.EnableDocking(CBRS_ALIGN_ANY);
	EnableDocking(CBRS_ALIGN_ANY);
	DockControlBar(&m_wndToolBar);

	NewScene();
	ASSERT(m_pScene);

	// Create a shape to add
	C3dShape* pShape = new C3dShape;
	pShape->CreateSphere(0.5);
	m_pScene->AddChild(pShape);
	m_pScene->m_ShapeList.Append(pShape);

	// Create the controller object
	if (m_iControlType == 0) {
		m_pController = new C3dPosCtlr;
	} else {
		m_pController = new C3dFlyCtlr;
	}
	m_pController->Create(&m_wnd3d,
						  OnGetCtrlFrame,
					      this);

	// restore the controller settings
	m_pController->SelectDevice(m_iObjCtrlDev);

	// enable mouse selection of objects
	m_wnd3d.EnableMouseSelection(OnSelChange, this);
	
	// show the current selection
	MakeCurrent(pShape);

	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	// Restore any saved settings
	RestoreSettings();

	// Restore the window size and position
	cs.x = m_rcWnd.left;
	cs.y = m_rcWnd.top;
	cs.cx = m_rcWnd.Width();
	cs.cy = m_rcWnd.Height();

	return CFrameWnd::PreCreateWindow(cs);
}

BOOL CMainFrame::NewScene()
{
	HideSelection();

	// Delete any scene we might have
	if (m_pScene) {
		m_wnd3d.SetScene(NULL);
		delete m_pScene;
		m_pScene = NULL;
	}
	MakeCurrent(NULL);

	// Create an initial scene
	m_pScene = new C3dScene;
	if (!m_pScene->Create()) return FALSE;

	// Set up the lighting
    C3dDirLight dl;
    dl.Create(0.8, 0.8, 0.8);
    m_pScene->AddChild(&dl);
    dl.SetPosition(-2, 2, -5);
    dl.SetDirection(1, -1, 1);
    m_pScene->SetAmbientLight(0.4, 0.4, 0.4);

	// reset the camera position and direction
    m_pScene->SetCameraPosition(C3dVector(0, 0, -10));
    m_pScene->SetCameraDirection(C3dVector(0, 0, 1), C3dVector(0, 1, 0));


	m_wnd3d.SetScene(m_pScene);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CFrameWnd::Dump(dc);
}

#endif //_DEBUG

void CMainFrame::OnDestroy() 
{
	CFrameWnd::OnDestroy();

	// Save the current settings
	SaveSettings();
	
	// remove any selection box
	HideSelection();

	// shut down any modeless dialogs
	if (m_pTransDlg) {
		m_pTransDlg->DestroyWindow();
	}

	// destroy the controller
	if (m_pController) {
		delete m_pController;
		m_pController = NULL;
	}

	// Delete any scene we might have
	m_wnd3d.SetScene(NULL);
	if (m_pScene) {
		delete m_pScene;
		m_pScene = NULL;
	}
	MakeCurrent(NULL);
}

/////////////////////////////////////////////////////////////////////////////
// window layout handler

void CMainFrame::RecalcLayout(BOOL bNotify) 
{
    // Rearrange the control bars and fit the 3D window in the middle
	// Let the frame rearrange the control bars
	CFrameWnd::RecalcLayout(bNotify);

	// Find the space that's left over
	CRect rc;
	RepositionBars(0,
				   0xFFFF,
				   IDC_3DWND,
				   CWnd::reposQuery,
				   &rc);
	if (IsWindow(m_wnd3d.GetSafeHwnd())) {
		m_wnd3d.MoveWindow(&rc, FALSE);
	}
}


///////////////////////////////////////////////////////////////
// Windows messages we need to handle

void CMainFrame::OnActivateApp(BOOL bActive, HTASK hTask) 
{
	CFrameWnd::OnActivateApp(bActive, hTask);
	
	// Tell the 3D window about the new state
	m_wnd3d.SendMessage(WM_ACTIVATEAPP,
						(WPARAM)bActive,
						(LPARAM)hTask);
}

void CMainFrame::OnPaletteChanged(CWnd* pFocusWnd) 
{
	// Let the 3D window know
	m_wnd3d.SendMessage(WM_PALETTECHANGED,
					   pFocusWnd ? (WPARAM)pFocusWnd->GetSafeHwnd() : 0);
}

void CMainFrame::OnMove(int x, int y) 
{
	CFrameWnd::OnMove(x, y);
	
	// Let the 3D window know the frame has moved
	m_wnd3d.SendMessage(WM_MOVE, 0, MAKELPARAM(0, 0));
}

///////////////////////////////////////////////////////////////
// Adding new objects

void CMainFrame::OnFileNew() 
{
	NewScene();
}

BOOL CMainFrame::OpenFile(const char* pszPath)
{
	// Create a new scene
	NewScene();
	ASSERT(m_pScene);

	// Try for a shape file
	C3dShape* pShape = new C3dShape;
	const char* pszFile = pShape->Load(pszPath);
	if (!pszFile) {
		delete pShape;
		return FALSE;
	}


	// Attach the new shape to the scene
	m_pScene->AddChild(pShape);

	// Add it to the shape list
	m_pScene->m_ShapeList.Append(pShape);

	// Make it the current shape
	MakeCurrent(pShape);

	// Add the name to the recent file list
	AfxGetApp()->AddToRecentFileList(pszFile);
	Status("Opened: %s", pszFile);
	return TRUE;
}

void CMainFrame::OnFileOpen() 
{
	OpenFile(NULL);
}

void CMainFrame::OnEditInsertShape() 
{
	CShapeDlg dlg("Insert Shape", this);

	if (dlg.DoModal() != IDOK) return;

	// Add the new shape
	if (dlg.m_pShape) {
		ASSERT(m_pScene);
		m_pScene->AddChild(dlg.m_pShape);
		m_pScene->m_ShapeList.Append(dlg.m_pShape);
		m_iHitFace = -1;
		MakeCurrent(dlg.m_pShape);
	}
}

///////////////////////////////////////////////////////////////
// Format a string for the status bar

void CMainFrame::Status(const char* pszFormat, ...)
{
	char buf[256];
	va_list arglist;
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);
	m_wndStatusBar.SetWindowText(buf);
	m_wndStatusBar.UpdateWindow();
}

/////////////////////////////////////////////////////////////
// Transform dialog

// callback from the dialog	when it closes
// static
void CMainFrame::OnCloseTrfmDlg(void* pArg)
{
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));
	pThis->m_pTransDlg = NULL;
}

// callback from the dialog to get a frame to control
// static
C3dFrame* CMainFrame::OnGetFrame(void* pArg)
{
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));
	return pThis->m_pCurShape;
}

void CMainFrame::ShowTrfmDlg()
{
	ASSERT(m_pCurShape);
	if (m_pTransDlg) {
		m_pTransDlg->BringWindowToTop();
	} else {
		m_pTransDlg = new CTransformDlg("Transforms",
										OnCloseTrfmDlg,
										OnGetFrame,
										this);
	}
}

void CMainFrame::OnEditTransform() 
{
	ShowTrfmDlg();
}

void CMainFrame::OnUpdateEditTransform(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pCurShape != NULL);
}

void CMainFrame::OnSetFocus(CWnd* pOldWnd) 
{
	CFrameWnd::OnSetFocus(pOldWnd);
	
	// Give the focus to the 3D window so it can process
	// keystrokes for the key input device
	m_wnd3d.SetFocus();
}

//////////////////////////////////////////////////////////////
// Control device selection

// callback from the dialog to get a frame to control
// static
C3dFrame* CMainFrame::OnGetCtrlFrame(void* pArg)
{
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));
	if (pThis->m_iControlObj == 1) {
		return pThis->m_wnd3d.GetCamera();
	}
	return pThis->m_pCurShape;
}

void CMainFrame::OnEditControl() 
{
	CInpDevDlg dlg;
	dlg.m_iObject = m_iControlObj;
	dlg.m_iType = m_iControlType;
	dlg.m_iDevice = m_iObjCtrlDev;
	if (dlg.DoModal() != IDOK) return;
	m_iControlObj = dlg.m_iObject;

	// set up the controller type
	if (dlg.m_iType != m_iControlType) {

		if (m_pController) delete m_pController;

		// create a new controller
		m_iControlType = dlg.m_iType;
		if (m_iControlType == 0) {
			m_pController = new C3dPosCtlr;
		} else {
			m_pController = new C3dFlyCtlr;
		}
		m_pController->Create(&m_wnd3d,
							  OnGetCtrlFrame,
							  this);
	}

	ASSERT(m_pController);
	m_pController->SelectDevice(dlg.m_iDevice);
 	m_iObjCtrlDev = m_pController->GetCurDevice();
	Status("Control is now via: %s", 
		   m_pController->GetDeviceName(m_iObjCtrlDev));
}

void CMainFrame::OnUpdateEditControl(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pController ? TRUE : FALSE);
}

void CMainFrame::OnEditInputdev() 
{
	ASSERT(m_pController);
	C3dInputDevice* pDev = m_pController->GetInputDevice();
	ASSERT(pDev);
	pDev->ShowConfigDlg();
}

//////////////////////////////////////////////////////////////////
// Save and Restore Settings

void CMainFrame::SaveSettings()
{
	CWinApp* pa = AfxGetApp();
	pa->WriteProfileInt("Settings","ControlDevice", m_iObjCtrlDev);
	CRect rc;
	GetWindowRect(&rc);
	pa->WriteProfileInt("Settings\\Window","Top", rc.top);
	pa->WriteProfileInt("Settings\\Window","Left", rc.left);
	pa->WriteProfileInt("Settings\\Window","Width", rc.Width());
	pa->WriteProfileInt("Settings\\Window","Height", rc.Height());
	pa->WriteProfileInt("Settings\\Options","ViewSel", m_bViewSelection);
	pa->WriteProfileInt("Settings\\Options","ViewHit", m_bViewHitPoint);
}

void CMainFrame::RestoreSettings()
{
	CWinApp* pa = AfxGetApp();
	m_iObjCtrlDev = pa->GetProfileInt("Settings","ControlDevice", 0);
	m_rcWnd.top = pa->GetProfileInt("Settings\\Window","Top", 50);
	m_rcWnd.left = pa->GetProfileInt("Settings\\Window","Left", 50);
	m_rcWnd.right = m_rcWnd.left + pa->GetProfileInt("Settings\\Window","Width", 300); 
	m_rcWnd.bottom = m_rcWnd.top + pa->GetProfileInt("Settings\\Window","Height", 350);
	m_bViewSelection = (BOOL) pa->GetProfileInt("Settings\\Options","ViewSel", 1);
	m_bViewHitPoint = (BOOL) pa->GetProfileInt("Settings\\Options","ViewHit", 1);
}

//////////////////////////////////////////////////////////////////////////
// object selection

// callback notification function
// called by m_wnd3d when the object selection changes
// static
void CMainFrame::OnSelChange(C3dShape* pShape, CPoint pt,
							 void* pArg)
{
	// Get a pointer to class
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));

	if (pShape) {

		// Make sure it's not a hit on selection box or hit-
		// pointer shapes
		if (!pShape->IsPartOf(pThis->m_pSelBox) 
		    && !pShape->IsPartOf(pThis->m_pHitPtr)) { 

			// Identify which face was hit
			C3dViewport* pViewport = pThis->m_wnd3d.GetStage()->GetViewport();
			pShape->HitTest(pt, pViewport,
				            &pThis->m_iHitFace,
							&pThis->m_vHitPoint);

		} else {

			pShape = NULL;

		}
	}

	// Make selected shape current
	pThis->MakeCurrent(pShape);
}

void CMainFrame::MakeCurrent(C3dShape* pShape)
{
	HideSelection();
	m_pCurShape = pShape;
	ShowSelection();
	if (m_pCurShape != NULL) {
		if (m_iHitFace >= 0) {
			Status("Selected: %s (%d @ %3.1f,%3.1f,%3.1f)",
				   m_pCurShape->GetName(),
				   m_iHitFace,
				   m_vHitPoint.x,
				   m_vHitPoint.y,
				   m_vHitPoint.z);
			m_pCurShape->SetFaceColor(m_iHitFace, 1, 0, 0);
		} else {
			Status("Selected: %s", m_pCurShape->GetName());
		}
	} else {
		Status("No selection");
	}
}

void CMainFrame::ShowSelection()
{
	// Remove any existing selection
	HideSelection();

	if (!m_pCurShape) {
		m_iHitFace = -1;
		return; // Nothing selected
	}

	if (m_bViewSelection) {

		// Get bounding box of shape
		double x1, x2, y1, y2, z1, z2;
		BOOL b = m_pCurShape->GetBox(x1, y1, z1, x2, y2,
			                         z2);
		ASSERT(b);

		// Create new selection box around shape
		m_pSelBox = new C3dShape;

		double r = 0.03;
		double rc = r * 2;
		C3dShape r1, r2, r3, r4, r5, r6, r7, r8, r9, r10, r11, r12, rd, ru, cd, cu;
		// Create rods for tubular frame
		r1.CreateRod(x1, y1, z1, x2, y1, z1, r);
		m_pSelBox->AddChild(&r1);
		r2.CreateRod(x1, y1, z1, x1, y2, z1, r);
		m_pSelBox->AddChild(&r2);
		r3.CreateRod(x1, y1, z1, x1, y1, z2, r);
		m_pSelBox->AddChild(&r3);
		r4.CreateRod(x1, y2, z1, x2, y2, z1, r);
		m_pSelBox->AddChild(&r4);
		r5.CreateRod(x2, y1, z1, x2, y2, z1, r);
		m_pSelBox->AddChild(&r5);
		r6.CreateRod(x1, y2, z1, x1, y2, z2, r);
		m_pSelBox->AddChild(&r6);
		r7.CreateRod(x2, y2, z1, x2, y2, z2, r);
		m_pSelBox->AddChild(&r7);
		r8.CreateRod(x2, y1, z1, x2, y1, z2, r);
		m_pSelBox->AddChild(&r8);
		r9.CreateRod(x1, y1, z2, x1, y2, z2, r);
		m_pSelBox->AddChild(&r9);
		r10.CreateRod(x1, y1, z2, x2, y1, z2, r);
		m_pSelBox->AddChild(&r10);
		r11.CreateRod(x2, y1, z2, x2, y2, z2, r);
		m_pSelBox->AddChild(&r11);
		r12.CreateRod(x1, y2, z2, x2, y2, z2, r);
		m_pSelBox->AddChild(&r12);
		//Create rods and cones for look and up
		rd.CreateRod(0, 0, 0, 0, 0, z2 * 1.2, r);
		m_pSelBox->AddChild(&rd);
		cd.CreateCone(0, 0, z2 * 1.2, rc, TRUE, 0, 0,
			          z2 * 1.4, 0, FALSE);
		m_pSelBox->AddChild(&cd);

		ru.CreateRod(0, 0, 0, 0, y2 * 1.1, 0, r);
		m_pSelBox->AddChild(&ru);
 		cu.CreateCone(0, y2 * 1.1, 0, rc, TRUE, 0,
			          y2 * 1.2, 0, 0, FALSE);
		m_pSelBox->AddChild(&cu);

		// Set position and orientation of box to
		// match shape
		double x, y, z, xd, yd, zd, xu, yu, zu;
		m_pCurShape->GetPosition(x, y, z);
		m_pCurShape->GetDirection(xd, yd, zd, xu, yu, zu);
		m_pSelBox->SetPosition(x, y, z);
		m_pSelBox->SetDirection(xd, yd, zd, xu, yu, zu);

		// Attach selection box to current shape
		// to move with shape
		m_pCurShape->AddChild(m_pSelBox);
	}

	if (m_bViewHitPoint && (m_iHitFace >= 0)) {

		// Get hit point on shape and
		// convert from local to world coordinates
		C3dVector vh = m_pCurShape->Transform(m_vHitPoint);

		// Get face normal
		ASSERT(m_pCurShape);
		C3dVector vn = m_pCurShape->GetFaceNormal(m_iHitFace);

		// Alter length as needed, add it to hit
		// point and convert to world coordinates
		C3dVector vx = vn * 0.5 + m_vHitPoint;
		vx = m_pCurShape->Transform(vx);

		// Point to the hit point with cone
		m_pHitPtr = new C3dShape;
		m_pHitPtr->CreateCone(vh.x, vh.y, vh.z, 0, FALSE,
							  vx.x, vx.y, vx.z, 0.1, TRUE);

		// Attach cone to the shape
		m_pCurShape->AddChild(m_pHitPtr);
	}
}

void CMainFrame::HideSelection()
{
	if (m_pSelBox) {
		ASSERT(m_pCurShape);
		m_pCurShape->RemoveChild(m_pSelBox);
		delete m_pSelBox;
		m_pSelBox = NULL;
	}
	if (m_pHitPtr) {
		ASSERT(m_pCurShape);
		m_pCurShape->RemoveChild(m_pHitPtr);
		delete m_pHitPtr;
		m_pHitPtr = NULL;
	}
}

void CMainFrame::OnUpdateViewSelection(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_bViewSelection);	
}

void CMainFrame::OnViewSelection() 
{
	m_bViewSelection = !m_bViewSelection;
	ShowSelection();
}

void CMainFrame::OnUpdateEditDeleteshape(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pCurShape != NULL);
}

void CMainFrame::OnEditDeleteshape() 
{
	ASSERT(m_pScene);
	ASSERT(m_pCurShape);
	C3dShape* pShape = m_pCurShape;
	MakeCurrent(NULL);
	m_pScene->RemoveChild(pShape);
	m_pScene->m_ShapeList.Delete(pShape);
}


void CMainFrame::OnUpdateViewHitpt(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_bViewHitPoint);	
}

void CMainFrame::OnViewHitpt() 
{
	m_bViewHitPoint = !m_bViewHitPoint;
	ShowSelection();
}
